import 'package:flutter/cupertino.dart';
import 'package:flutter_easyrefresh/easy_refresh.dart';
import 'package:flutter_mylibrary/base/provider/request_info.dart';
import 'package:flutter_mylibrary/base/provider/view_state_model.dart';
import 'package:flutter_mylibrary/config/const.dart';
import 'package:flutter_mylibrary/flutter_mylibrary.dart';
import 'package:flutter_mylibrary/http/dio_utils.dart';
import 'package:flutter_mylibrary/http/exception/http_error.dart';
import 'package:flutter_mylibrary/utils/utils.dart';
import 'package:flutter_mylibrary/widget/widget_template.dart';


/// 基于
abstract class ViewStateRefreshListModel<T> extends ViewStateModel {

  ///自定义第一页页数
  int firstPage() => startPage;

  ///自定义一页数量
  int pageSize() => pageLimit;

  RequestInfo getRequestInfo();

  /// 页面数据
  List<T> list = [];

  onCompleted(List<T> data) {}


  ///获取数据
  Future<List<T>> getData() {
    RequestInfo requestInfo = getRequestInfo();
    return DioUtils.getInstance().request(
        requestInfo?.method,
        requestInfo?.url,
        data: requestInfo?.data,
        publicMethod: false,
        onFinish: (){
          refreshController.resetRefreshState();
          refreshController.resetLoadState();
        },
        onError: (e){
          handleError(e);
        }
    ).then((res){
      try{
        List<T> resData = res.data is List? List<T>.from(res.data.map((e) => requestInfo?.handleRes(e))).toList() : requestInfo?.handleRes(res.data);
        if (isNotEmpty(resData) && resData.length < pageSize()) {
          ///提前预判
          hasMore = false;
        }
        if (isEmpty(resData) && isEmpty(list)) {
          ///空数据
          setEmpty();
        }
        list.addAll(resData);
        setIdle();
        refreshController.finishRefresh(success: true);
        refreshController.finishLoad(success: true, noMore: !hasMore);
        return resData;
      }catch(e){
        handleError(e);
        return null;
      }
    });
  }


  void handleError(e){
    if (hasMore) {
      _currentPageNum--;
    }
    refreshController.finishRefresh(success: false);
    refreshController.finishLoad(success: false, noMore: !hasMore);
    if (!isInit) {
      if (e is HttpError) {
        setError(HttpError(HttpError.UNKNOWN, e.toString()), e.toString());
      }else{
        setError(e, e.message, message: e.message);
      }
    }
  }


  void initData({bool retry = false}){
    if (retry) {
      setBusy();
    }
    refresh();
  }



  ///获取一个公共的刷新控件(EasyRefresh)
  Widget getEasyRefreshWidget({ViewStateModelWidgetListBuilder<List<T>> widgetListBuilder, ViewStateModelWidgetBuilder<List<T>> builder, Widget emptyWidget, Widget busyWidget, bool showErrorWidget = true, bool showEmptyWidget = true, bool showBusyWidget = true}){
    Widget emptyErrorWidget = getRefreshNoDataWidget(viewStateModel: this, busyWidget: busyWidget, emptyDataWidget: emptyWidget, showErrorWidget: showErrorWidget, showEmptyWidget: showEmptyWidget, showBusyWidget: showBusyWidget);
    return EasyRefresh.custom(
      emptyWidget: emptyErrorWidget,
      slivers: widgetListBuilder(list)??[
        isEmpty(emptyErrorWidget)? builder(list) : SliverToBoxAdapter(child: Container(),),
      ],
      onRefresh: isBusy? null : () async{
        await refresh();
      },
      onLoad: (isBusy || !hasMore)? null : () async{
        await loadMore();
      },
      controller: refreshController,
    );
  }





  /// 当前页码
  int _currentPageNum;

  int get currentPageNum => _currentPageNum;


  /// 下拉刷新
  ///
  /// [init] 是否是第一次加载
  /// true:  Error时,需要跳转页面
  /// false: Error时,不需要跳转页面,直接给出提示
  Future<List<T>> refresh() async {
    list.clear();
    if (getRequestInfo().useLoadMore) {
      hasMore = true;
    }
    _currentPageNum = firstPage();
    var data = await getData();
    onCompleted(data);
    return data;
  }

  /// 上拉加载更多
  Future<List<T>> loadMore() async {
    _currentPageNum = ++_currentPageNum;
    var data = await getData();
    onCompleted(data);
    return data;
  }

  bool _hasMore;

  ///设置是否还有更多
  set hasMore(bool hasMore){
    this._hasMore = hasMore;
  }

  bool get hasMore => getRequestInfo().useLoadMore?_hasMore??true : false;

  @override
  void dispose() {
    refreshController.dispose();
    super.dispose();
  }


}
